@@ -26,6 +26,7 @@
 #include "gok-button.h"
 #include "gok-keyboard.h"
 #include "gok-log.h"
+#include "gok-data.h"
 #include "gok-scanner.h"
 #include "main.h"
 #include <stdio.h>
@@ -38,6 +39,8 @@ GokKey* gok_button_find_key (GtkWidget* pWidget);
 
 static GtkToggleButtonClass *parent_class = NULL;
 
+void gok_input_detach_corepointer (void);
+
 /* TODO: make image/text relationship (i.e. image on left, label on right) configurable */
 
 /**
@@ -215,6 +218,10 @@ gint gok_button_enter_notify   (GtkWidget *widget, GdkEventCrossing   *event)
 	pKey = gok_button_find_key (widget);
 	if (pKey != NULL)
 	{
+		if (!strcmp (gok_data_get_name_accessmethod (), "dwellselection")) 
+		{
+		    gok_main_warn_corepointer (FALSE, FALSE, FALSE);
+		}
 		gok_scanner_on_key_enter (pKey);
 	}
 
@@ -304,6 +311,7 @@ GokKey* gok_button_find_key (GtkWidget* pWidget)
 	return NULL;
 }
 
+static gboolean
 gok_button_expose (GtkWidget      *widget,
 		   GdkEventExpose *event)
 {
