@@ -21,7 +21,7 @@
 
 #include <gnome.h>
 
-gboolean gok_settings_page_keysizespace_initialize (GtkWidget* pWindowSettings);
+gboolean gok_settings_page_keysizespace_initialize (GladeXML* xml);
 void gok_settings_page_keysizespace_refresh (void);
 gboolean gok_settings_page_keysizespace_ok (void);
 gboolean gok_settings_page_keysizespace_revert (void);
