@@ -133,7 +133,15 @@ static gulong keyboardmods;
 static gboolean spyshutdown = FALSE;/* this global could be removed if we have a
 						way of removing our idle handler in gok_spy_stop */
 
-#define MAX_CHILD_COUNT 100
+/* set later using gconf key values  */
+static gint gok_max_search_depth = 18;
+static gint gok_max_search_breadth = 100;
+/* define the upper and lower ranges of values these values */
+#define GOK_MAX_SEARCH_DEPTH 25
+#define GOK_MIN_SEARCH_DEPTH 5
+#define GOK_MAX_SEARCH_BREADTH 25
+#define GOK_MIN_SEARCH_BREADTH 5
+
 #define MAX_NAME_SEARCH_DEPTH 10
 
 static void
@@ -305,8 +313,31 @@ gok_spy_idle_handler (gpointer data)
 void gok_spy_open()
 {
 	gboolean success = FALSE;
+	GConfClient *gconf_client;
 
 	gok_log_enter ();
+	/* read in gconf search values */
+	gconf_client = gconf_client_get_default ();
+	gok_gconf_get_int ( gconf_client, "/apps/gok/spy/gui_search_depth", &gok_max_search_depth);
+	/* i18n note: TODO mark these for translation after freeze */
+	if (gok_max_search_depth > GOK_MAX_SEARCH_DEPTH)  {
+		g_warning ("gconf key gok_max_search_depth exceded, using %d>",GOK_MAX_SEARCH_DEPTH);
+		gok_max_search_depth = GOK_MAX_SEARCH_DEPTH;
+	}
+	else if (gok_max_search_depth < GOK_MIN_SEARCH_DEPTH) {
+		g_warning ("gconf key gok_max_search_depth too small, using %d>",GOK_MIN_SEARCH_DEPTH);
+		gok_max_search_depth = GOK_MIN_SEARCH_DEPTH;
+	}
+	gok_gconf_get_int ( gconf_client, "/apps/gok/spy/gui_search_breadth", &gok_max_search_breadth);
+	if (gok_max_search_depth > GOK_MAX_SEARCH_BREADTH)  {
+		g_warning ("gconf key gok_max_search_breadth exceded, using %d>",GOK_MAX_SEARCH_BREADTH);
+		gok_max_search_breadth = GOK_MAX_SEARCH_BREADTH;
+	}
+	else if (gok_max_search_depth < GOK_MIN_SEARCH_BREADTH) {
+		g_warning ("gconf key gok_max_search_breadth too low, using %d>",GOK_MIN_SEARCH_BREADTH);
+		gok_max_search_breadth = GOK_MIN_SEARCH_BREADTH;
+	}
+	
 	if (m_gokSpyOpen != TRUE)
 	{
 		eventQueue = g_queue_new();
@@ -796,6 +827,7 @@ gok_spy_get_table_nodes (Accessible *search_root)
 	g_assert (Accessible_isTable (search_root));
 	table = Accessible_getTable (search_root);
 	flags.value = ~0;
+	flags.data.inside_table = TRUE;
 	flags.data.inside_html_container = FALSE;
 	col_count = AccessibleTable_getNColumns (table);
 	if (col_count > 1)
@@ -856,7 +888,7 @@ gok_spy_get_table_nodes (Accessible *search_root)
 			if (name) SPI_freeString (name);
 			if (desc) SPI_freeString (desc);
 			/* we make exception for compound cells, which will scavenge a name from their children */
-			if (has_name || has_desc || (Accessible_getChildCount (child) > 0))
+			if (has_name || has_desc || (Accessible_getChildCount (child) > 0) || Accessible_isAction (child))
 			{
 				nodes = gok_spy_append_node (nodes, child, flags);
 				gok_spy_accessible_unref (child);
@@ -913,7 +945,7 @@ GSList *
 gok_spy_get_actionable_descendants (Accessible *accessible, GSList *nodes)
 {
     gint child_count, i, max_children = 20;
-    Accessible *child;
+    Accessible *child, *parent;
     AccessibleNodeFlags flags;
     GSList *ret_nodes;
     gboolean is_selection = Accessible_isSelection (accessible);
@@ -921,6 +953,13 @@ gok_spy_get_actionable_descendants (Accessible *accessible, GSList *nodes)
     flags.data.inside_html_container = FALSE;
     flags.data.is_ui = TRUE;
 
+    parent = Accessible_getParent (accessible);
+    if (parent) {
+	if (Accessible_isTable (parent))
+	    flags.data.inside_table = TRUE;
+	Accessible_unref (parent);
+    }
+
     if (Accessible_isAction (accessible))
     {
 	ret_nodes = gok_spy_append_node (nodes, accessible, flags);
@@ -1824,6 +1863,126 @@ GSList* gok_spy_append_node (GSList* nodes,
 			}
 		}
 	}
+
+	/* use row and column info inside a table */
+	if (pName == NULL && flags.data.inside_table) {
+		long iip, irow = -1, icol = -1;
+		AccessibleTable* table;
+		Accessible *parent, *col, *row;
+
+		parent = Accessible_getParent (pAccessible);
+		if (Accessible_isTable (parent)) {
+			iip = Accessible_getIndexInParent (pAccessible);
+			table = Accessible_getTable (parent);
+			if (table) {
+			    irow = AccessibleTable_getRowAtIndex (table, iip);
+			    icol = AccessibleTable_getColumnAtIndex (table, iip);
+			}
+			if (irow >= 0 && icol > 0) {
+			        gchar *rowstring = NULL;
+				gchar *colstring = NULL;
+				/* row */
+				row = AccessibleTable_getRowHeader (table, irow);
+				if (row != NULL) {
+					rowstring = Accessible_getName (row);
+					Accessible_unref (row);
+				}
+				if (rowstring && strlen (rowstring)) {
+				    /* use header name */
+				    pName = g_strdup (rowstring);
+				}
+				else {
+				    /* try scavenging first column of this row */
+				    gchar *firstcellname = NULL;
+				    Accessible *cell = AccessibleTable_getAccessibleAt (table, irow, 0);
+				    if (cell) {
+					    firstcellname = Accessible_getName (cell);
+					    if (!firstcellname || !strlen (firstcellname)) {
+						if (firstcellname) {
+						    SPI_freeString (firstcellname);
+						    firstcellname = NULL;
+						}
+						if (Accessible_isText (cell)) {
+						    AccessibleText *celltext = Accessible_getText (cell);
+						    firstcellname = AccessibleText_getText (celltext, 0, -1);
+						    AccessibleText_unref (celltext);
+						}
+					    }
+					    if (!firstcellname || !strlen (firstcellname)) {
+						Accessible *cellchild;
+						gint childcount = Accessible_getChildCount (cell);
+						if (firstcellname) {
+						    SPI_freeString (firstcellname);
+						    firstcellname = NULL;
+						}
+						for (j = 0; j < 2 && j < childcount; ++j) {
+						    cellchild = Accessible_getChildAtIndex (cell, j);
+						    firstcellname = Accessible_getName (cellchild);
+						    if (!firstcellname || !strlen (firstcellname)) {
+							if (firstcellname) {
+							    SPI_freeString (firstcellname);
+							    firstcellname = NULL;
+							}
+							if (Accessible_isText (cellchild)) {
+							    AccessibleText *celltext = Accessible_getText (cellchild);
+							    gint count = AccessibleText_getCharacterCount (celltext);
+							    firstcellname = AccessibleText_getText (celltext, 0, MIN (count, 15));
+							    AccessibleText_unref (celltext);
+							}
+						    }
+						    Accessible_unref (cellchild);
+						    if (firstcellname && strlen (firstcellname)) break;
+						    else if (firstcellname) {
+							SPI_freeString (firstcellname);
+							firstcellname = NULL;
+						    }
+						}
+					    }
+					    Accessible_unref (cell);
+				    }
+				    if (firstcellname && strlen (firstcellname)) {
+					pName = g_strdup (firstcellname);
+					SPI_freeString (firstcellname);
+				    }
+				    else {
+					if (firstcellname) {
+					    SPI_freeString (firstcellname);
+					    firstcellname = NULL;
+					}
+					/* use row index */
+					pName = g_strdup_printf ("%d",irow);
+				    }
+				}
+				if (rowstring) SPI_freeString (rowstring);
+				col = AccessibleTable_getColumnHeader (table, icol);
+				if (col != NULL) {
+					colstring = Accessible_getName (col);
+					Accessible_unref (col);
+				}
+				if (colstring) {
+				    gchar* free_string = pName;
+				    if (strlen (colstring)) {
+					/* use header name */
+					pName = g_strconcat (pName, " ", colstring, NULL);
+					g_free (free_string);
+				    }
+				    else
+					SPI_freeString (colstring);
+				}
+				else {
+				    gchar *free_string = pName;
+				    /* use col index */
+				    pName = g_strconcat (pName, " ", g_strdup_printf ("%d",icol), NULL);
+				    g_free (free_string);
+				}
+				AccessibleTable_unref (table);
+			}
+		}
+		Accessible_unref (parent);
+		if (pName != NULL) {
+			nodes = gok_spy_add_node (nodes, pAccessible, flags, 0, pName);
+		}
+	}
 	
 	gok_log_leave();
 
@@ -2045,6 +2204,7 @@ void gok_spy_process_focus (Accessible* accessible)
 	Accessible* pnewWindowAccessible;
 	Accessible* ptempAccessible;
 	gboolean has_menu = FALSE;
+	AccessibleRole role;
 	
 	gok_log_enter();
 	g_assert(accessible != NULL);
@@ -2103,12 +2263,18 @@ void gok_spy_process_focus (Accessible* accessible)
 	 * if current active window is NULL, and this is a menu item, 
 	 * then we're in a popup menu of some kind; branch to the menus kbd
 	 */
+	role = Accessible_getRole (accessible);
 
 	if (accessible && (m_ptheWindowAccessible == NULL) && 
-	    gok_spy_is_menu_role (Accessible_getRole (accessible)))
+	    gok_spy_is_menu_role (role))
 	{
 	    AccessibleNode *node = g_new0 (AccessibleNode, 1);
-	    gchar *name = Accessible_getName (accessible);
+	    gchar *name;
+	    if (role == SPI_ROLE_MENU_ITEM)
+	    {
+		accessible = Accessible_getParent (accessible);
+	    }
+	    name = Accessible_getName (accessible);
 	    node->link = -1;
 	    node->flags.value = 0;
 	    node->flags.data.is_menu = TRUE;
@@ -2327,7 +2493,7 @@ gok_spy_find_and_append_toolbar_items (Accessible *root, AccessibleNodeFlags fla
 }
 
 static GokSpyUIFlags 
-gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboard_ui_flags, AccessibleNodeFlags context_flags)
+gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboard_ui_flags, AccessibleNodeFlags context_flags, gint search_depth)
 {
 	Accessible* child;
 	Accessible* parent;
@@ -2342,6 +2508,13 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	ui_flags.value = 0;
 	parent = rootAccessible;
 
+	if (++search_depth > gok_max_search_depth) 
+	{
+		g_warning ("GOK: maximum search depth exceeded.");
+		gok_log_leave ();
+		return ui_flags;
+	}			
+
 	if (gok_spy_worth_searching (parent) == FALSE) 
 	{
 		gok_log ("not worth searching.");
@@ -2356,9 +2529,9 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 			context_flags.data.inside_html_container = TRUE;
 	}
 	
-	/* MAX_CHILD_COUNT is a bit of a kludge and will cause certain table items to be unreachable via UI Grab */
-	if (child_count > MAX_CHILD_COUNT)
-		child_count = MAX_CHILD_COUNT;
+	/* gok_max_search_breadth is a bit of a kludge and will cause certain table items to be unreachable via UI Grab */
+	if (child_count > gok_max_search_breadth)
+		child_count = gok_max_search_breadth;
 	for (i = 0; i < child_count; ++i)
 	{
 		AccessibleNodeFlags flags;  /* TODO? memcpy context_flags? */
@@ -2366,6 +2539,9 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 		flags.data.inside_html_container = context_flags.data.inside_html_container;
 
 		child = Accessible_getChildAtIndex (parent, i);
+		if (child == parent) /* bogus hierarchy, but yelp sometimes does this */
+		    continue;
+
 		if (child)
 		{
 			gok_spy_accessible_implicit_ref (child);
@@ -2388,7 +2564,7 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				ui_flags.data.toolbars = flags.data.is_toolbar_item = TRUE;
 				if (!gok_spy_find_and_append_toolbar_items (child, flags)) 
 				{
-				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+				    ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
 				}
 			}
 			if ((keyboard_ui_flags.data.gui || keyboard_ui_flags.data.editable_text) && 
@@ -2414,12 +2590,12 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 				    /* special case for notebook tabs and expanders, whose children should also appear */
   				    if (role == SPI_ROLE_PAGE_TAB || role == SPI_ROLE_TOGGLE_BUTTON)
   				    {
-  					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+  					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
   				    }
 				}
 				else
 				{
-					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags).value;
+					ui_flags.value |= gok_spy_search_component_list (child, keyboard_ui_flags, context_flags, search_depth).value;
 				}
 			}
 			gok_spy_accessible_unref (child);
@@ -2461,7 +2637,8 @@ gok_spy_update_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 	_priv_ui_nodes = NULL;
 
 	context_flags.value = 0;
-	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags, context_flags);
+	ui_flags = gok_spy_search_component_list (rootAccessible, keyboard_ui_flags, context_flags, 0);
+
 	/* _priv_ui_nodes is now re-assigned, by above call. */
 	gok_spy_resolve_namesakes (_priv_ui_nodes);
 
